import json
from typing import Union

from attrs import asdict

from unofficial_shipengine.utils.serialize import serializer
from .models import ShipmentRequest, Shipment
from ..common.services import BaseService


class ShipmentService(BaseService):
    """
    ShipmentService provides methods for interacting with shipment-related endpoints in the ShipEngine API.

    Methods:
        create_shipment(shipment_request: Union[ShipmentRequest, list[ShipmentRequest]]) -> Union[Shipment, list[Shipment]]:
            Creates a shipment or a list of shipments.

        get_by_id(shipment_id: str) -> Shipment:
            Retrieves a shipment by its ID.

        get_by_external_id(external_shipment_id: str) -> Shipment:
            Retrieves a shipment by its external ID.

        update_shipment(shipment: Shipment) -> Shipment:
            Updates an existing shipment.

        cancel_shipment(shipment: Union[Shipment, str]) -> None:
            Cancels a shipment.
    """

    def create_shipment(
        self, shipment_request: Union[ShipmentRequest, list[ShipmentRequest]]
    ) -> Union[Shipment, list[Shipment]]:
        """
        Creates a shipment or a list of shipments.

        Args:
            shipment_request (Union[ShipmentRequest, list[ShipmentRequest]]): The shipment request or a list of shipment requests.

        Returns:
            Union[Shipment, list[Shipment]]: The created shipment or a list of created shipments.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(shipment_request, list):
            shipment_requests = shipment_request
        else:
            shipment_requests = [shipment_request]

        url = "https://api.shipengine.com/v1/shipments"
        data = [asdict(sr, value_serializer=serializer) for sr in shipment_requests]
        json_data: str = json.dumps({"shipments": data})

        response = self.session.post(url, data=json_data)
        self._handle_response(response)
        response_dict = response.json()

        shipments = [Shipment.from_dict(s) for s in response_dict["shipments"]]

        if isinstance(shipment_request, ShipmentRequest):
            return shipments[0]

        return shipments

    def get_by_id(self, shipment_id: str) -> Shipment:
        """
        Retrieves a shipment by its ID.

        Args:
            shipment_id (str): The ID of the shipment to retrieve.

        Returns:
            Shipment: The retrieved Shipment object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = f"https://api.shipengine.com/v1/shipments/{shipment_id}"

        response = self.session.get(url)
        self._handle_response(response)
        response_dict = response.json()

        return Shipment.from_dict(response_dict)

    def get_by_external_id(self, external_shipment_id: str) -> Shipment:
        """
        Retrieves a shipment by its external ID.

        Args:
            external_shipment_id (str): The external ID of the shipment to retrieve.

        Returns:
            Shipment: The retrieved Shipment object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = (
            f"https://api.shipengine.com/v1/shipments/"
            f"external_shipment_id/{external_shipment_id}"
        )

        response = self.session.get(url)
        self._handle_response(response)
        response_dict = response.json()

        return Shipment.from_dict(response_dict)

    def update_shipment(self, shipment: Shipment) -> Shipment:
        """
        Updates an existing shipment.

        Args:
            shipment (Shipment): The shipment object to update.

        Returns:
            Shipment: The updated Shipment object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = f"https://api.shipengine.com/v1/shipments/{shipment.shipment_id}"
        json_data = json.dumps(asdict(shipment, value_serializer=serializer))

        response = self.session.put(url, data=json_data)
        self._handle_response(response)
        response_dict = response.json()

        return Shipment.from_dict(response_dict)

    def cancel_shipment(self, shipment: Union[Shipment, str]) -> None:
        """
        Cancels a shipment.

        Args:
            shipment (Union[Shipment, str]): The shipment object or shipment ID to cancel.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(shipment, Shipment):
            shipment = shipment.shipment_id

        url = f"https://api.shipengine.com/v1/shipments/{shipment}/cancel"
        response = self.session.put(url)
        self._handle_response(response)



class ShipEngineAPIError(Exception):
    """Base class for all ShipEngine API errors."""

    def __init__(self, request_id: str, errors: list[str]):
        self.request_id = request_id
        self.errors = errors
        super().__init__(self._format_message())

    def _format_message(self):
        return f"ShipEngine API Error (Request ID: {self.request_id}): {self.errors}"



import json
from typing import Union

from attrs import asdict

from unofficial_shipengine.utils.serialize import serializer
from .models import WarehouseRequest, Warehouse
from ..common.services import BaseService


class WarehouseService(BaseService):
    """
    WarehouseService provides methods for interacting with warehouse-related endpoints in the ShipEngine API.

    Methods:
        create_warehouse(warehouse_request: WarehouseRequest) -> Warehouse:
            Creates a warehouse.

        delete_warehouse(warehouse: Union[Warehouse, str]) -> None:
            Deletes a warehouse.

        get_by_id(warehouse_id: str) -> Warehouse:
            Retrieves a warehouse by its ID.
    """

    def create_warehouse(self, warehouse_request: WarehouseRequest) -> Warehouse:
        """
        Create a new warehouse.

        Args:
            warehouse_request (WarehouseRequest): An object representing the details of the warehouse to be created.

        Returns:
            Warehouse: An object representing the newly created warehouse.

        Raises:
            ShipEngineAPIError: If the request to create the warehouse fails.
        """
        data: str = json.dumps(asdict(warehouse_request, value_serializer=serializer))
        url = "https://api.shipengine.com/v1/warehouses"
        response = self.session.post(url, data=data)
        response_dict = json.loads(response.text)
        self._handle_response(response)

        return Warehouse.from_dict(response_dict)

    def delete_warehouse(self, warehouse: Union[Warehouse, str]) -> None:
        """
        Delete a warehouse.

        Args:
            warehouse (Union[Warehouse, str]): The warehouse object or ID of the warehouse to be deleted.

        Raises:
            ShipEngineAPIError: If the request to delete the warehouse fails.
        """
        if isinstance(warehouse, Warehouse):
            warehouse = warehouse.warehouse_id

        url = f"https://api.shipengine.com/v1/warehouses/{warehouse}"
        response = self.session.delete(url)
        self._handle_response(response)

    def get_by_id(self, warehouse_id: str) -> Warehouse:
        """
        Retrieve a warehouse by its ID.

        Args:
            warehouse_id (str): The ID of the warehouse to retrieve.

        Returns:
            Warehouse: An object representing the retrieved warehouse.

        Raises:
            ShipEngineAPIError: If the request to retrieve the warehouse fails.
        """
        url = f"https://api.shipengine.com/v1/warehouses/{warehouse_id}"
        response = self.session.get(url)
        response_dict = json.loads(response.text)
        self._handle_response(response)

        warehouse: Warehouse = Warehouse.from_dict(response_dict)

        return warehouse



from typing import Self, Any

from attrs import define, field


@define
class UnofficialShipEngineConfig:
    """
    UnofficialShipEngineConfig stores the configuration for the UnofficialShipEngine client.

    Attributes:
        api_key (str): The API key for authenticating with the ShipEngine API.
        retries (int): The number of retries for API requests in case of failure. Defaults to 3.
        backoff_factor (float): The backoff factor for retrying API requests. Defaults to 0.5.
    """

    api_key: str
    retries: int = field(default=3)
    backoff_factor: float = field(default=0.5)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)



from enum import Enum
from typing import Union, Any, Self

from attrs import define, field, validators

from .enums import LabelLayout, LabelFormat
from ..common.models import URL, LabelDownload, Error
from ..labels.enums import DisplayScheme


@define
class ProcessLabels:
    create_batch_and_process_labels: bool = field(default=True)
    ship_date: str = field(default=None)

    label_layout: LabelLayout = field(
        default=LabelLayout.FOUR_BY_SIX, validator=validators.in_(LabelLayout)
    )

    label_format: LabelFormat = field(default=LabelFormat.PDF)

    display_scheme: DisplayScheme = field(
        default=DisplayScheme.LABEL, validator=validators.in_(DisplayScheme)
    )


@define
class BatchRequest:
    shipment_ids: list[str]
    rate_ids: list[str] = field(default=None)
    external_batch_id: str = field(default=None)
    batch_notes: str = field(default=None)
    process_labels: ProcessLabels = field(default=None)


@define
class Batch:
    class Status(Enum):
        OPEN: str = "open"
        QUEUED: str = "queued"
        PROCESSING: str = "processing"
        COMPLETED: str = "completed"
        COMPLETED_WITH_ERRORS: str = "completed_with_errors"
        ARCHIVED: str = "archived"
        NOTIFYING: str = "notifying"
        INVALID: str = "invalid"

    batch_id: str
    batch_number: str
    external_batch_id: str
    created_at: str
    processed_at: str
    errors: int
    process_errors: list[Error]
    warnings: int
    completed: int
    forms: int
    count: int
    batch_shipments_url: URL
    batch_labels_url: URL
    batch_errors_url: URL
    label_download: LabelDownload
    form_download: URL
    status: Status = field(validator=validators.in_(Status))
    paperless_download: URL = field(default=None)
    batch_notes: str = field(default=None)

    label_layout: Union[LabelLayout, None] = field(default=LabelLayout.FOUR_BY_SIX)
    label_format: LabelFormat = field(
        default=LabelFormat.PDF, validator=validators.in_(LabelFormat)
    )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        batch_shipments_url: URL = URL(**data.pop("batch_shipments_url"))
        batch_labels_url: URL = URL(**data.pop("batch_labels_url"))
        batch_errors_url: URL = URL(**data.pop("batch_errors_url"))
        label_download: LabelDownload = LabelDownload(**data.pop("label_download"))
        form_download: URL = URL(**data.pop("form_download"))

        return cls(
            batch_shipments_url=batch_shipments_url,
            batch_labels_url=batch_labels_url,
            batch_errors_url=batch_errors_url,
            label_download=label_download,
            form_download=form_download,
            **data,
        )



from typing import Self, Any

from attrs import define


@define
class CarrierOption:
    name: str
    default_value: str
    description: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)


@define
class CarrierPackage:
    package_id: str
    package_code: str
    name: str
    description: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)


@define
class CarrierService:
    carrier_id: str
    carrier_code: str
    service_code: str
    name: str
    domestic: bool
    international: bool
    is_multi_package_supported: bool

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)


@define
class CarrierBalance:
    currency: str
    amount: float

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(**data)


@define
class Carrier:
    carrier_id: str
    carrier_code: str
    account_number: str
    requires_funded_amount: bool
    balance: float
    nickname: str
    friendly_name: str
    primary: bool
    has_multi_package_supporting_services: bool
    supports_label_messages: str
    disabled_by_billing_plan: bool
    funding_source_id: str
    packages: list[CarrierPackage] = []
    services: list[CarrierService] = []
    options: list[CarrierOption] = []

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        packages: list[CarrierPackage] = [
            CarrierPackage.from_dict(p) for p in data.pop("packages")
        ]

        services: list[CarrierService] = [
            CarrierService.from_dict(s) for s in data.pop("services")
        ]

        options: list[CarrierOption] = [
            CarrierOption.from_dict(o) for o in data.pop("options")
        ]

        return cls(packages=packages, services=services, options=options, **data)



import json
from typing import Union, Optional, Any

from attrs import asdict

from unofficial_shipengine.utils.serialize import serializer
from .models import Batch, BatchRequest, ProcessLabels
from ..common.services import BaseService
from ..shipments.models import Shipment


class BatchService(BaseService):
    """
    BatchService provides methods for interacting with batch-related endpoints in the ShipEngine API.

    Methods:
        create_batch(batch_request: BatchRequest) -> Batch:
            Creates a new batch based on the given batch request.

        get_by_id(batch_id: str) -> Batch:
            Retrieves a batch by its ID.

        process_labels(batch: Union[Batch, str], process_labels: ProcessLabels) -> None:
            Processes labels for the given batch.

        get_batch_errors(batch: Union[Batch, str], page: int = 1, pagesize: int = 1) -> dict[str, Any]:
            Retrieves errors for the given batch.

        delete_batch(batch: Union[Batch, str]) -> None:
            Deletes the given batch.

        add_to_batch(batch: Union[Batch, str], shipments: list[Union[Shipment, str]], rates: Optional[list[str]] = None) -> None:
            Adds shipments to the given batch.

        remove_from_batch(batch: Union[Batch, str], shipments: list[Union[Shipment, str]], rates: Optional[list[str]] = None) -> None:
            Removes shipments from the given batch.
    """

    def create_batch(self, batch_request: BatchRequest) -> Batch:
        """
        Creates a new batch based on the given batch request.

        Args:
            batch_request (BatchRequest): The request data for creating a batch.

        Returns:
            Batch: The created batch object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        data: str = json.dumps(asdict(batch_request, value_serializer=serializer))

        response = self.session.post("https://api.shipengine.com/v1/batches", data=data)
        response_dict = json.loads(response.text)
        self._handle_response(response)

        return Batch.from_dict(response_dict)

    def get_by_id(self, batch_id: str) -> Batch:
        """
        Retrieves a batch by its ID.

        Args:
            batch_id (str): The ID of the batch to retrieve.

        Returns:
            Batch: The retrieved batch object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url: str = f"https://api.shipengine.com/v1/batches/{batch_id}"

        response = self.session.get(url)
        response_dict = json.loads(response.text)
        self._handle_response(response)

        return Batch.from_dict(response_dict)

    def process_labels(
        self, batch: Union[Batch, str], process_labels: ProcessLabels
    ) -> None:
        """Function can take a Batch object or a batch_id."""
        if isinstance(batch, Batch):
            batch = batch.batch_id

        data: str = json.dumps(asdict(process_labels, value_serializer=serializer))
        url: str = f"https://api.shipengine.com/v1/batches/{batch}/process/labels"
        response = self.session.post(url, data=data)
        self._handle_response(response)

    def get_batch_errors(
        self, batch: Union[Batch, str], page: int = 1, pagesize: int = 1
    ) -> dict[str, Any]:
        """
        Retrieves errors for the given batch.

        Args:
            batch (Union[Batch, str]): The batch object or batch ID.
            page (int, optional): The page number to retrieve. Defaults to 1.
            pagesize (int, optional): The number of errors per page. Defaults to 1.

        Returns:
            dict[str, Any]: The errors for the batch.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(batch, Batch):
            batch = batch.batch_id

        url: str = f"https://api.shipengine.com/v1/batches/{batch}/errors"
        response = self.session.get(
            url, params=json.dumps({"page": page, "pagesize": pagesize})
        )

        response_json: dict[str, Any] = response.json()

        return response_json

    def delete_batch(self, batch: Union[Batch, str]) -> None:
        """
        Deletes the given batch.

        Args:
            batch (Union[Batch, str]): The batch object or batch ID.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(batch, Batch):
            batch = batch.batch_id

        url: str = f"https://api.shipengine.com/v1/batches/{batch}"
        response = self.session.delete(url)
        self._handle_response(response)

    def add_to_batch(
        self,
        batch: Union[Batch, str],
        shipments: list[Union[Shipment, str]],
        rates: Optional[list[str]] = None,
    ) -> None:
        """
        Adds shipments to the given batch.

        Args:
            batch (Union[Batch, str]): The batch object or batch ID.
            shipments (list[Union[Shipment, str]]): A list of shipment objects or shipment IDs.
            rates (Optional[list[str]], optional): A list of rate IDs. Defaults to None.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        self._modify_batch(batch, "add", shipments, rates)

    def remove_from_batch(
        self,
        batch: Union[Batch, str],
        shipments: list[Union[Shipment, str]],
        rates: Optional[list[str]] = None,
    ) -> None:
        """
        Removes shipments from the given batch.

        Args:
            batch (Union[Batch, str]): The batch object or batch ID.
            shipments (list[Union[Shipment, str]]): A list of shipment objects or shipment IDs.
            rates (Optional[list[str]], optional): A list of rate IDs. Defaults to None.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        self._modify_batch(batch, "remove", shipments, rates)

    def _modify_batch(
        self,
        batch: Union[Batch, str],
        endpoint: str,
        shipments: list[Union[str, Shipment]],
        rates: Optional[list[str]] = None,
    ) -> None:
        if isinstance(batch, Batch):
            batch = batch.batch_id

        shipments = [s if isinstance(s, str) else s.shipment_id for s in shipments]

        url: str = f"https://api.shipengine.com/v1/batches/{batch}/{endpoint}"
        data: str = json.dumps({"shipment_ids": shipments, "rate_ids": rates})
        response = self.session.post(url, data=data)
        self._handle_response(response)



from enum import Enum


class Confirmation(Enum):
    NONE: str = "none"
    DELIVERY: str = "delivery"
    SIGNATURE: str = "signature"
    ADULT_SIGNATURE: str = "adult_signature"
    DIRECT_SIGNATURE: str = "direct_signature"
    DELIVERY_MAILED: str = "delivery_mailed"
    VERBAL_CONFIRMATION: str = "verbal_confirmation"


class InsuranceProvider(Enum):
    NONE: str = "none"
    SHIPSURANCE: str = "shipsurance"
    CARRIER: str = "carrier"
    THIRD_PARTY: str = "third_party"


class OrderSourceCode(Enum):
    AMAZON_CA: str = "amazon_ca"
    AMAZON_US: str = "amazon_us"
    BRIGHTPEARL: str = "brightpearl"
    CHANNEL_ADVISOR: str = "channel_advisor"
    CRATEJOY: str = "cratejoy"
    EBAY: str = "ebay"
    ETSY: str = "etsy"
    JANE: str = "jane"
    GROUPON_GOODS: str = "groupon_goods"
    MAGENTO: str = "magento"
    PAYPAL: str = "paypal"
    SELLER_ACTIVE: str = "seller_active"
    SHOPIFY: str = "shopify"
    STITCH_LABS: str = "stitch_labs"
    SQUARESPACE: str = "squarespace"
    THREE_DCART: str = "three_dcart"
    TOPHATTER: str = "tophatter"
    WALMART: str = "walmart"
    WOO_COMMERCE: str = "woo_commerce"
    VOLUSION: str = "volusion"



from unofficial_shipengine.core.tracking.models import TrackingInformation
from ..common.services import BaseService


class TrackingService(BaseService):
    """
    TrackingService provides methods for interacting with tracking-related endpoints in the ShipEngine API.

    Methods:
        get_tracking_information(carrier_code: str, tracking_number: str) -> TrackingInformation:
            Retrieves tracking information for a package.

        start_tracking_package(carrier_code: str, tracking_number: str) -> None:
            Starts tracking a package.

        stop_tracking_package(carrier_code: str, tracking_number: str) -> None:
            Stops tracking a package.
    """

    def get_tracking_information(
        self, carrier_code: str, tracking_number: str
    ) -> TrackingInformation:
        """
        Retrieves tracking information for a package.

        Args:
            carrier_code (str): The carrier code.
            tracking_number (str): The tracking number.

        Returns:
            TrackingInformation: The tracking information for the package.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = "https://api.shipengine.com/v1/tracking"
        params = {"carrier_code": carrier_code, "tracking_number": tracking_number}

        response = self.session.get(url, params=params)
        self._handle_response(response)
        response_dict = response.json()

        tracking_information: TrackingInformation = TrackingInformation.from_dict(
            response_dict
        )

        return tracking_information

    def start_tracking_package(self, carrier_code: str, tracking_number: str) -> None:
        """
        Starts tracking a package.

        Args:
            carrier_code (str): The carrier code.
            tracking_number (str): The tracking number.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        self._track_package("start", carrier_code, tracking_number)

    def stop_tracking_package(self, carrier_code: str, tracking_number: str) -> None:
        """
        Stops tracking a package.

        Args:
            carrier_code (str): The carrier code.
            tracking_number (str): The tracking number.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        self._track_package("stop", carrier_code, tracking_number)

    def _track_package(
        self, action: str, carrier_code: str, tracking_number: str
    ) -> None:
        """
        Internal method to start or stop tracking a package.

        Args:
            action (str): The action to perform ("start" or "stop").
            carrier_code (str): The carrier code.
            tracking_number (str): The tracking number.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = f"https://api.shipengine.com/v1/tracking/{action}"
        params = {"carrier_code": carrier_code, "tracking_number": tracking_number}
        response = self.session.post(url, params=params)
        self._handle_response(response)



import json
from typing import Union

from attrs import asdict

from unofficial_shipengine.utils.serialize import serializer
from .models import Label, LabelRequest, ReturnLabelRequest
from ..common.services import BaseService
from ..shipments.models import Shipment
from ..tracking.models import TrackingInformation


class LabelService(BaseService):
    """
    LabelService provides methods for interacting with label-related endpoints in the ShipEngine API.

    Methods:
        purchase_label(label_request: LabelRequest) -> Label:
            Purchases a shipping label.

        create_return_label(label: Union[Label, str], return_label_request: ReturnLabelRequest) -> Label:
            Creates a return label for an existing label.

        get_by_id(label_id: str) -> Label:
            Retrieves a label by its ID.

        get_label_tracking_info(label: Union[Label, str]) -> TrackingInformation:
            Retrieves tracking information for a label.
    """

    def purchase_label(self, label_request: LabelRequest) -> Label:
        """
        Purchases a shipping label.

        Args:
            label_request (LabelRequest): The request data for purchasing a label.

        Returns:
            Label: The purchased Label object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = "https://api.shipengine.com/v1/labels"
        json_data = json.dumps(asdict(label_request, value_serializer=serializer))

        response = self.session.post(url, data=json_data)
        self._handle_response(response)
        response_dict = response.json()

        label: Label = Label.from_dict(response_dict)

        return label

    def create_return_label(
        self, label: Union[Label, str], return_label_request: ReturnLabelRequest
    ) -> Label:
        """
        Creates a return label for an existing label.

        Args:
            label (Union[Label, str]): The original label object or label ID.
            return_label_request (ReturnLabelRequest): The request data for creating a return label.

        Returns:
            Label: The created return Label object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(label, Label):
            label = label.label_id

        url = f"https://api.shipengine.com/v1/labels/{label}/return"
        json_data = json.dumps(
            asdict(return_label_request, value_serializer=serializer)
        )

        response = self.session.post(url, data=json_data)
        self._handle_response(response)
        response_dict = response.json()

        return_label: Label = Label.from_dict(response_dict)

        return return_label

    def get_by_id(self, label_id: str) -> Label:
        """
        Retrieves a label by its ID.

        Args:
            label_id (str): The ID of the label to retrieve.

        Returns:
            Label: The retrieved Label object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = f"https://api.shipengine.com/v1/labels/{label_id}"
        response = self.session.get(url)
        self._handle_response(response)
        response_dict = response.json()

        label: Label = Label.from_dict(response_dict)

        return label

    def get_label_tracking_info(self, label: Union[Label, str]) -> TrackingInformation:
        """
        Retrieves tracking information for a label.

        Args:
            label (Union[Label, str]): The label object or label ID.

        Returns:
            TrackingInformation: The tracking information for the label.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(label, Label):
            label = label.label_id

        url = f"https://api.shipengine.com/v1/labels/{label}/track"
        response = self.session.get(url)
        self._handle_response(response)
        response_dict = response.json()

        tracking_information: TrackingInformation = TrackingInformation.from_dict(
            response_dict
        )

        return tracking_information



from enum import Enum
from typing import Self, Any

from attrs import define, field, validators

from .enums import Confirmation, InsuranceProvider, OrderSourceCode
from ..common.enums import ValidateAddress
from ..common.models import (
    Value,
    Address,
    Weight,
    AddressValidation,
    Package,
)


@define
class TaxIdentifier:
    class TaxableEntityType(Enum):
        SHIPPER: str = "shipper"
        RECIPIENT: str = "recipient"
        IOR: str = "ior"

    class IdentifierType(Enum):
        VAT: str = "vat"
        EORI: str = "eori"
        SSN: str = "ssn"
        EIN: str = "ein"
        TIN: str = "tin"
        IOSS: str = "ioss"
        PAN: str = "pan"
        VOEC: str = "voec"
        PCCC: str = "pccc"
        OSS: str = "oss"
        PASSPORT: str = "passport"
        ABN: str = "abn"

    taxable_entity_type: TaxableEntityType = field(
        validator=validators.in_(TaxableEntityType)
    )
    identifier_type: IdentifierType = field(validator=validators.in_(IdentifierType))
    issuing_authority: str
    value: str


@define
class CustomsInformation:
    @define
    class InvoiceAdditionalDetail:
        freight_charge: Value
        insurance_charge: Value
        discount: Value
        other_charge: Value
        other_charge_description: str = field(default=None)

        @other_charge_description.validator
        def _validate_other_charge(self, attribute, value):
            if self.other_charge is not None and value is None:
                raise ValueError(
                    "other_charge_description required when other_charge is set"
                )

    class Contents(Enum):
        MERCHANDISE: str = "merchandise"
        DOCUMENTS: str = "documents"
        GIFT: str = "gift"
        RETURNED_GOODS: str = "returned_goods"
        SAMPLE: str = "sample"
        OTHER: str = "other"

    class NonDelivery(Enum):
        RETURN_TO_SENDER: str = "return_to_sender"
        TREAT_AS_ABANDONED: str = "treat_as_abandoned"

    class TermsOfTradeCode(Enum):
        EXW: str = "exw"
        FCA: str = "fca"
        CPT: str = "cpt"
        CIP: str = "cip"
        DPU: str = "dpu"
        DAP: str = "dap"
        DDP: str = "ddp"
        FAS: str = "fas"
        FOB: str = "fob"
        CFR: str = "fr"
        CIF: str = "cif"
        DDU: str = "ddu"
        DAF: str = "daf"
        DEQ: str = "deq"
        DES: str = "des"

    declaration: str
    invoice_additional_details: InvoiceAdditionalDetail
    importer_of_record: Address
    terms_of_trade_code: TermsOfTradeCode = field(
        validator=validators.in_(TermsOfTradeCode)
    )
    contents: Contents = field(
        default=Contents.MERCHANDISE, validator=validators.in_(Contents)
    )
    contents_explanation: str = field(default=None)
    non_delivery: NonDelivery = field(
        default=NonDelivery.RETURN_TO_SENDER, validator=validators.in_(NonDelivery)
    )

    @contents_explanation.validator
    def _validate_contents_explanation(self, attribute, value):
        if self.contents == self.Contents.OTHER and value is None:
            raise ValueError(
                "'contents_explanation' is required when 'contents' are 'other'"
            )


@define
class AdvancedOptions:
    @define
    class FedexFreight:
        shipper_load_and_count: str
        booking_confirmation: str

    @define
    class CollectOnDelivery:
        class PaymentType(Enum):
            ANY: str = "any"
            CASH: str = "cash"
            CASH_EQUIVALENT: str = "cash_equivalent"
            NONE: str = "none"

        payment_type: PaymentType = field(validator=validators.in_(PaymentType))
        payment_amount: Value

    @define
    class DangerousGoodsContact:
        name: str
        phone: str

    class BillToParty(Enum):
        RECIPIENT: str = "recipient"
        THIRD_PARTY: str = "third_party"

    class OriginType(Enum):
        PICKUP: str = "pickup"
        DROP_OFF: str = "drop_off"

    contains_alcohol: bool = field(default=False)
    delivered_duty_paid: bool = field(default=False)
    dry_ice: bool = field(default=False)
    dry_ice_weight: Weight = field(default=None)
    non_machinable: bool = field(default=False)
    saturday_delivery: bool = field(default=False)
    bill_to_account: str = field(default=None)
    bill_to_country_code: str = field(default=None)
    bill_to_party: BillToParty = field(default=None)
    bill_to_postal_code: str = field(default=None)
    fedex_freight: FedexFreight = field(default=None)
    use_ups_ground_freight_pricing: bool = field(default=None)
    freight_class: str = field(default=None)
    custom_field1: str = field(default=None)
    custom_field2: str = field(default=None)
    custom_field3: str = field(default=None)
    origin_type: OriginType = field(default=None)
    additional_handling: bool = field(default=None)
    shipper_release: bool = field(default=None)
    collect_on_delivery: CollectOnDelivery = field(default=None)
    third_party_consignee: bool = field(default=False)
    dangerous_goods: bool = field(default=False)
    dangerous_goods_contact: DangerousGoodsContact = field(default=None)

    ancillary_endorsements_option: str = field(default=None)
    return_pickup_attempts: int = field(default=None)
    own_document_upload: bool = field(default=False)
    limited_quantity: bool = field(default=False)
    event_notification: bool = field(default=False)

    @use_ups_ground_freight_pricing.validator
    def _use_ups_ground_freight_pricing(self, attribute, value):
        if value is not None and self.freight_class is None:
            raise ValueError(
                "freight_class cannot be None when "
                "use_ups_ground_freight_pricing is set"
            )


@define
class ShipmentRequest:
    carrier_id: str
    service_code: str
    ship_to: Address
    ship_date: str = field(default=None)
    validate_address: ValidateAddress = field(
        default=ValidateAddress.NO_VALIDATION, validator=validators.in_(ValidateAddress)
    )
    advanced_options: AdvancedOptions = field(default=None)
    confirmation: Confirmation = field(
        default=Confirmation.NONE, validator=validators.in_(Confirmation)
    )
    tags: list[str] = field(default=[])
    is_return: bool = field(default=False)
    customs: list[CustomsInformation] = field(default=None)
    warehouse_id: str = field(default=None)
    ship_from: Address = field(default=None)
    return_to: Address = field(default=None)
    items: list[str] = field(default=[])
    external_order_id: str = field(default=None)
    tax_identifiers: list[TaxIdentifier] = field(default=None)
    external_shipment_id: str = field(default=None)
    shipment_number: str = field(default=None)
    insurance_provider: InsuranceProvider = field(
        default=InsuranceProvider.NONE, validator=validators.in_(InsuranceProvider)
    )
    order_source_code: OrderSourceCode = field(default=None)
    packages: list[Package] = field(default=None)
    comparison_rate_type: str = field(default=None)


@define(kw_only=True)
class Shipment(ShipmentRequest):
    """Subclassed from ShipmentRequest to avoid repeating many fields."""

    class Status(Enum):
        PENDING: str = "pending"
        PROCESSING: str = "processing"
        LABEL_PURCHASED: str = "label_purchased"
        CANCELLED: str = "cancelled"

    shipment_id: str
    created_at: str
    modified_at: str
    shipping_rule_id: str
    errors: list[str] = field(default=[])
    shipment_status: Status = field(
        default=Status.PENDING, validator=validators.in_(Status)
    )
    total_weight: Weight = field(default=None)
    address_validation: AddressValidation = field(default=None)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        ship_to = Address(**data.pop("ship_to"))
        ship_from = Address(**data.pop("ship_from"))
        return_to = Address(**data.pop("return_to"))
        packages = [Package.from_dict(p) for p in data.pop("packages")]
        total_weight = Weight(**data.pop("total_weight"))
        advanced_options = AdvancedOptions(**data.pop("advanced_options"))

        return cls(
            ship_to=ship_to,
            ship_from=ship_from,
            advanced_options=advanced_options,
            return_to=return_to,
            packages=packages,
            total_weight=total_weight,
            **data,
        )



from enum import Enum


class ChargeEvent(Enum):
    CARRIER_DEFAULT: str = "carrier_default"
    ON_CREATION: str = "on_creation"
    ON_CARRIER_ACCEPTANCE: str = "on_carrier_acceptance"


class LabelFormat(Enum):
    PDF: str = "pdf"
    PNG: str = "png"
    ZPL: str = "zpl"


class LabelLayout(Enum):
    FOUR_BY_SIX: str = "4x6"
    LETTER: str = "letter"


class DisplayScheme(Enum):
    LABEL: str = "label"
    QR_CODE: str = "qr_code"
    LABEL_AND_QR_CODE: str = "label_and_qr_code"
    PAPERLESS: str = "paperless"
    LABEL_AND_PAPERLESS: str = "label_and_paperless"


class LabelDownloadType(Enum):
    URL: str = "url"
    INLINE: str = "inline"



from typing import Self, Any

from attrs import define, field

from ..common.models import Address


@define
class WarehouseRequest:
    name: str
    origin_address: Address
    return_address: Address = field(default=None)
    is_default: bool = field(default=None)


@define
class Warehouse:
    warehouse_id: str
    name: str
    created_at: str
    origin_address: Address
    return_address: Address
    is_default: bool = field(default=None)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        origin_address: Address = Address(**data.pop("origin_address"))
        return_address: Address = Address(**data.pop("return_address"))
        return cls(origin_address=origin_address, return_address=return_address, **data)



from typing import Any

import requests

from unofficial_shipengine.exceptions import ShipEngineAPIError


class BaseService:
    def __init__(self, session: requests.Session):
        self.session = session

    @staticmethod
    def _handle_response(response: requests.Response):
        if response.status_code in [200, 204, 207]:
            return

        response_dict: dict[str, Any] = response.json()

        raise ShipEngineAPIError(
            request_id=response_dict["request_id"], errors=response_dict["errors"]
        )



from typing import Union, Mapping, Any

import requests
from requests.adapters import HTTPAdapter, Retry

from .core.batches.services import BatchService
from .core.carriers.services import CarrierService
from .core.labels.services import LabelService
from .core.shipments.services import ShipmentService
from .core.tracking.services import TrackingService
from .core.warehouses.services import WarehouseService
from .unofficial_shipengine_config import UnofficialShipEngineConfig


class UnofficialShipEngine:
    def __init__(
        self,
        config: Union[
            UnofficialShipEngineConfig, dict[str, Union[float, int, str]], str
        ],
    ) -> None:
        """
        Initializes the UnofficialShipEngine client with the provided configuration.

        Args:
            config (Union[UnofficialShipEngineConfig, dict[str, Union[float, int, str]], str]):
                The configuration for the ShipEngine client. It can be an UnofficialShipEngineConfig object,
                a dictionary, or a string representing just the API key and the rest of the values will default.
        """
        self.config = self._parse_config(config)
        self._session = self._create_session()

        self.shipments = ShipmentService(self._session)
        self.carriers = CarrierService(self._session)
        self.batches = BatchService(self._session)
        self.warehouses = WarehouseService(self._session)
        self.labels = LabelService(self._session)
        self.tracking = TrackingService(self._session)

    @staticmethod
    def _parse_config(
        config: Union[UnofficialShipEngineConfig, dict[str, Any], str],
    ) -> UnofficialShipEngineConfig:
        """
        Parses the provided configuration into an UnofficialShipEngineConfig object.

        Args:
            config (Union[UnofficialShipEngineConfig, dict[str, Any], str]):
                The configuration to parse. It can be an UnofficialShipEngineConfig object,
                a dictionary, or a string representing just the API key and the rest of the values will default.

        Returns:
            UnofficialShipEngineConfig: The parsed configuration object.

        Raises:
            ValueError: If the configuration type is invalid.
        """
        if isinstance(config, str):
            return UnofficialShipEngineConfig(config)
        elif isinstance(config, Mapping):
            return UnofficialShipEngineConfig.from_dict(config)
        elif isinstance(config, UnofficialShipEngineConfig):
            return config
        else:
            raise ValueError("Invalid configuration type provided")

    def _create_session(self) -> requests.Session:
        """
        Creates a configured requests session for making API calls.

        Returns:
            requests.Session: The configured session with headers and retry strategy.
        """
        session = requests.Session()
        session.headers = {
            "Host": "api.shipengine.com",
            "API-Key": self.config.api_key,
            "Content-Type": "application/json",
        }

        retry = Retry(
            total=self.config.retries,
            backoff_factor=self.config.backoff_factor,
            status_forcelist=[500, 502, 503, 504],
        )

        session.mount("https://", HTTPAdapter(max_retries=retry))

        return session



import json
from typing import Union

from .models import Carrier, CarrierBalance
from ..common.services import BaseService


class CarrierService(BaseService):
    """
    CarrierService provides methods for interacting with carrier-related endpoints in the ShipEngine API.

    Methods:
        get_carriers() -> list[Carrier]:
            Retrieves a list of carriers.

        get_by_id(carrier_id: str) -> Carrier:
            Retrieves a carrier by its ID.

        add_funds(carrier: Union[Carrier, str], amount: float, currency: str = "usd") -> CarrierBalance:
            Adds funds to a carrier account.
    """

    def get_carriers(self) -> list[Carrier]:
        """
        Retrieves a list of carriers.

        Returns:
            list[Carrier]: A list of Carrier objects.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = "https://api.shipengine.com/v1/carriers"
        response = self.session.get(url)
        response_dict = json.loads(response.text)

        self._handle_response(response)

        carriers = response_dict["carriers"]

        return [Carrier.from_dict(c) for c in carriers]

    def get_by_id(self, carrier_id: str) -> Carrier:
        """
        Retrieves a carrier by its ID.

        Args:
            carrier_id (str): The ID of the carrier to retrieve.

        Returns:
            Carrier: The retrieved Carrier object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        url = f"https://api.shipengine.com/v1/carriers/{carrier_id}"
        response = self.session.get(url)
        response_dict = json.loads(response.text)

        self._handle_response(response)

        return Carrier.from_dict(response_dict)

    def add_funds(
        self, carrier: Union[Carrier, str], amount: float, currency: str = "usd"
    ) -> CarrierBalance:
        """
        Adds funds to a carrier account.

        Note: There is no test mode for adding funds. You will be charged when you add funds.
              You can either pass a Carrier object or just the carrier_id as a string.

        Args:
            carrier (Union[Carrier, str]): The carrier object or carrier ID to add funds to.
            amount (float): The amount of funds to add.
            currency (str, optional): The currency in which to add funds. Defaults to "usd".

        Returns:
            CarrierBalance: The updated CarrierBalance object.

        Raises:
            ShipEngineAPIError: If the response from the API is invalid.
        """
        if isinstance(carrier, Carrier):
            carrier = carrier.carrier_id

        url = f"https://api.shipengine.com/v1/carriers/{carrier}/add_funds"
        data = {"amount": amount, "currency": currency}

        response = self.session.post(url, data=json.dumps(data))
        response_dict = json.loads(response.text)

        self._handle_response(response)

        return CarrierBalance.from_dict(response_dict)



from enum import Enum
from typing import Any, Self, Optional

from attrs import define, field, validators

from .enums import (
    ChargeEvent,
    LabelFormat,
    LabelLayout,
    DisplayScheme,
    LabelDownloadType,
)
from ..common.enums import ValidateAddress
from ..common.models import Value, URL, Package
from ..shipments.models import ShipmentRequest


@define
class LabelDownload:
    href: str
    pdf: str = field(default=None)
    zpl: str = field(default=None)
    png: str = field(default=None)


@define
class PaperlessDownload:
    href: str
    instructions: str = field(default=None)
    handoff_code: str = field(default=None)


@define
class ReturnLabelRequest:
    charge_event: ChargeEvent = field(
        default=ChargeEvent.CARRIER_DEFAULT, validator=validators.in_(ChargeEvent)
    )
    label_layout: LabelLayout = field(
        default=LabelLayout.FOUR_BY_SIX, validator=validators.in_(LabelLayout)
    )
    display_scheme: DisplayScheme = field(
        default=DisplayScheme.LABEL, validator=validators.in_(DisplayScheme)
    )
    label_format: LabelFormat = field(
        default=LabelFormat.PDF, validator=validators.in_(LabelFormat)
    )
    label_download_type: LabelDownloadType = field(
        default=LabelDownloadType.URL, validator=validators.in_(LabelDownloadType)
    )
    label_image_id: str = field(default=None)


@define
class PackageLabel(Package):
    tracking_number: Optional[str] = field(default=None)
    qr_code_download: Optional[str] = field(default=None)
    paperless_download: str = field(default=None)
    sequence: Optional[str] = field(default=None)
    alternative_identifiers: Optional[list[str]] = field(default=None)
    has_label_documents: bool = field(default=None)
    has_form_documents: bool = field(default=None)
    has_qr_code_documents: bool = field(default=None)
    has_paperless_label_documents: bool = field(default=None)


@define
class LabelRequest:
    shipment: ShipmentRequest
    label_image_id: str = field(default=None)
    label_layout: LabelLayout = field(
        default=LabelLayout.FOUR_BY_SIX, validator=validators.in_(LabelLayout)
    )
    display_scheme: DisplayScheme = field(
        default=DisplayScheme.LABEL, validator=validators.in_(DisplayScheme)
    )
    label_format: LabelFormat = field(
        default=LabelFormat.PDF, validator=validators.in_(LabelFormat)
    )
    label_download_type: LabelDownloadType = field(
        default=LabelDownloadType.URL, validator=validators.in_(LabelDownloadType)
    )
    validate_address: ValidateAddress = field(
        default=ValidateAddress.NO_VALIDATION, validator=validators.in_(ValidateAddress)
    )
    outbound_label_id: str = field(default=None)
    charge_event: ChargeEvent = field(
        default=ChargeEvent.CARRIER_DEFAULT, validator=validators.in_(ChargeEvent)
    )
    is_return_label: bool = field(default=None)
    rma_number: str = field(default=None)
    ship_to_service_point_id: str = field(default=None)
    ship_form_service_point_id: str = field(default=None)


@define
class Label:
    @define
    class AlternativeIdentifier:
        type: str
        value: str

    class Status(Enum):
        PROCESSING: str = "processing"
        COMPLETED: str = "completed"
        ERROR: str = "error"
        VOIDED: str = "voided"

    class TrackingStatus(Enum):
        UNKNOWN: str = "unknown"
        IN_TRANSIT: str = "in_transit"
        ERROR: str = "error"
        DELIVERED: str = "delivered"

    label_id: str
    status: Status = field(validator=validators.in_(Status))
    shipment_id: str
    ship_date: str
    created_at: str
    shipment_cost: Value
    insurance_cost: Value
    requested_comparison_amount: Value
    tracking_number: str
    is_return_label: bool
    is_international: bool
    carrier_id: str
    charge_event: ChargeEvent
    service_code: str
    voided: bool
    voided_at: str
    trackable: bool
    carrier_code: str
    tracking_status: TrackingStatus = field(validator=validators.in_(TrackingStatus))
    label_download: LabelDownload
    form_download: URL
    paperless_download: PaperlessDownload
    insurance_claim: URL
    packages: list[PackageLabel]
    alternative_identifiers: list[AlternativeIdentifier] = field(default=None)
    label_format: LabelFormat = field(
        default=LabelFormat.PDF, validator=validators.in_(LabelFormat)
    )
    display_scheme: DisplayScheme = field(
        default=DisplayScheme.LABEL, validator=validators.in_(DisplayScheme)
    )
    label_layout: LabelLayout = field(
        default=LabelLayout.FOUR_BY_SIX, validator=validators.in_(LabelLayout)
    )
    rate_details: list[str] = field(default=None)
    label_image_id: str = field(default=None)
    batch_id: str = field(default=None)
    rma_number: str = field(default=None)
    package_code: str = field(default=None)
    qr_code_download: URL = field(default=None)
    shipping_rule_id: str = field(default=None)

    @property
    def total_cost(self) -> Value:
        return Value(
            currency=self.shipment_cost.currency,
            amount=self.shipment_cost.amount + self.insurance_cost.amount,
        )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        if "form_download" in data.keys() and data["form_download"] is not None:
            form_download: URL = URL(**data.pop("form_download"))
        else:
            form_download = data.pop("form_download", None)

        if (
            "paperless_download" in data.keys()
            and data["paperless_download"] is not None
        ):
            paperless_download: PaperlessDownload = PaperlessDownload(
                **data.pop("paperless_download")
            )
        else:
            paperless_download = data.pop("paperless_download", None)

        label_download: LabelDownload = LabelDownload(**data.pop("label_download"))
        shipment_cost: Value = Value(**data.pop("shipment_cost"))
        insurance_cost: Value = Value(**data.pop("insurance_cost"))
        requested_comparison_amount: Value = Value(
            **data.pop("requested_comparison_amount")
        )

        packages: list[PackageLabel] = [
            PackageLabel.from_dict(p) for p in data.pop("packages")
        ]

        return cls(
            label_download=label_download,
            form_download=form_download,
            shipment_cost=shipment_cost,
            insurance_cost=insurance_cost,
            requested_comparison_amount=requested_comparison_amount,
            paperless_download=paperless_download,
            packages=packages,
            **data
        )



from enum import Enum
from typing import Self, Optional, Any

from attrs import define, field, validators


@define
class Address:
    class AddressResidentialIndicator(Enum):
        UNKNOWN: str = "unknown"
        YES: str = "yes"
        NO: str = "no"

    name: str
    phone: str
    address_line1: str
    city_locality: str
    state_province: str
    postal_code: str
    address_line2: str = field(default=None)
    address_line3: str = field(default=None)
    email: str = field(default=None)
    company_name: str = field(default=None)
    instructions: str = field(default=None)
    geolocation: list[object] = field(default=None)
    country_code: str = field(default="US")
    address_residential_indicator: AddressResidentialIndicator = field(
        default=AddressResidentialIndicator.UNKNOWN
    )


@define
class Error:
    error_source: str
    error_type: str
    error_code: str
    error_message: str


@define
class URL:
    href: str
    type: str = field(default=None)


@define
class LabelDownload:
    href: str
    pdf: str = field(default=None)
    zpl: str = field(default=None)
    png: str = field(default=None)


@define
class AddressValidation:
    @define
    class Message:
        code: str
        message: str
        type: str
        detail_code: str = field(default=None)

    class Status(Enum):
        UNVERIFIED: str = "unverified"
        VERIFIED: str = "verified"
        WARNING: str = "warning"
        ERROR: str = "error"

    status: Status
    original_address: Address
    matched_address: Address = field(default=None)
    messages: list[Message] = field(default=None)


@define
class Weight:
    class Unit(Enum):
        POUND = "pound"
        OUNCE = "ounce"
        GRAM = "gram"
        KILOGRAM = "kilogram"

    value: float
    unit: Unit = field(validator=validators.in_(Unit))


@define
class Dimension:
    class Unit(Enum):
        INCH = "inch"
        CENTIMETER = "centimeter"

    length: float
    width: float
    height: float
    unit: Unit = field(default=Unit.INCH, validator=validators.in_(Unit))


@define
class Value:
    # ISO 4217: https://www.iso.org/iso-4217-currency-codes.html
    currency: str = field(default="usd")
    amount: float = field(default=0.0)


@define
class GenericAmount:
    amount: int = field(default=0)
    unit: str = field(default=None)


@define
class Product:

    @define
    class DangerousGood:

        class PackagingGroup(Enum):
            I: str = "i"
            II: str = "ii"
            III: str = "iii"

        class PackagingInstructionSection(Enum):
            SECTION_1: str = "section_1"
            SECTION_2: str = "section_2"
            SECTION_1A: str = "section_1a"
            SECTION_1B: str = "section_1b"

        class TransportMean(Enum):
            GROUND: str = "ground"
            WATER: str = "water"
            CARGO_AIRCRAFT_ONLY: str = "cargo_aircraft_only"
            PASSENGER_AIRCRAFT: str = "passenger_aircraft"

        class RegulationLevel(Enum):
            LIGHLTY_REGULATED: str = "lightly_regulated"
            FULLY_REGULATED: str = "fully_regulated"
            LIMITED_QUANTITIES: str = "limited_quantities"
            EXCEPTED_QUANTITY: str = "excepted_quantity"

        dangerous_amount: GenericAmount
        quantity: int = field(default=0)
        packaging_instruction: str = field(default=None)
        id_number: str = field(default=None)
        shipping_name: str = field(default=None)
        technical_name: str = field(default=None)
        product_class: str = field(default=None)
        product_class_subsidiary: str = field(default=None)
        packaging_type: str = field(default=None)
        transport_category: str = field(default=None)
        regulation_authority: str = field(default=None)
        radioactive: bool = field(default=None)
        reportable_quantity: bool = field(default=None)
        tunnel_code: str = field(default=None)
        additional_description: str = field(default=None)
        regulation_level: RegulationLevel = field(default=None)
        transport_mean: TransportMean = field(default=None)
        packaging_group: PackagingGroup = field(default=None)
        packaging_instruction_section: PackagingInstructionSection = field(default=None)

    quantity: int
    value: Value
    weight: Weight
    sku: str = field(default=None)
    sku_description: str = field(default=None)
    mid_code: str = field(default=None)
    product_url: str = field(default=None)
    vat_rate: float = field(default=None)
    unit_of_measure: str = field(default=None)
    country_of_origin: str = field(default=None)
    harmonized_tarrif_code: str = field(default=None)
    description: str = field(default=None)


@define
class Package:
    @define
    class LabelMessages:
        reference1: Optional[str] = field(default=None)
        reference2: Optional[str] = field(default=None)
        reference3: Optional[str] = field(default=None)

    weight: Weight
    package_code: Optional[str] = field(default=None)
    dimensions: Optional["Dimension"] = field(default=None)
    content_description: Optional[str] = field(default=None)
    package_id: Optional[str] = field(default=None)
    insured_value: Optional["Value"] = field(default=None)
    label_messages: Optional[LabelMessages] = field(default=None)
    products: Optional[list[Product]] = field(default=None)
    external_package_id: Optional[str] = field(default=None)
    shipment_package_id: Optional[str] = field(default=None)
    package_name: Optional[str] = field(default=None)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        weight = Weight(**data.pop("weight"))
        dimensions = Dimension(**data.pop("dimensions"))
        insured_value = Value(**data.pop("insured_value"))
        label_messages = Package.LabelMessages(**data.pop("label_messages"))

        return cls(
            weight=weight,
            dimensions=dimensions,
            insured_value=insured_value,
            label_messages=label_messages,
            **data
        )



from typing import Any, Self

from attrs import define

from .enums import TrackingStatusCode


@define
class TrackEvent:
    occurred_at: str
    carrier_occurred_at: str
    description: str
    city_locality: str
    state_province: str
    postal_code: str
    country_code: str
    company_name: str
    signer: str
    event_code: str
    carrier_detail_code: str
    status_code: TrackingStatusCode
    status_description: str
    carrier_status_code: str
    carrier_status_description: str
    latitude: float
    longitude: float


@define
class TrackingInformation:

    tracking_number: str
    tracking_url: str
    status_code: TrackingStatusCode
    carrier_code: str
    carrier_id: str
    status_description: str
    carrier_status_code: str
    carrier_detail_code: str
    carrier_status_description: str
    ship_date: str
    estimated_delivery_date: str
    actual_delivery_date: str
    exception_description: str
    events: list[TrackEvent]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        events: list[TrackEvent] = [TrackEvent(**e) for e in data.pop("events")]
        return cls(events=events, **data)
